(* jpp.ml -- Java pretty printer.  Prints TaML code into corresponding Java *)

open Ast
open Sast
open Type

let default_keyword = "public static"	(*put before  global variables and functions*)
let default_implicit_param = "null"		(*for null declaration*)

(*translate binary operators*)
let string_of_operator = function
	 	Add 		-> "+"    	| Sub -> "-"  | Mult -> "*"      	| Div -> "/"
  | Equal -> "==" 			| Neq -> "!=" | Less_than -> "<" 	| Leq -> "<=" 
	| Greater_than -> ">" | Geq -> ">=" |	Or ->"||"   			|	And -> "&&"
	|	Mod 		->"%"
	
(*translate unary operators*)	
let string_of_operator1 = function
	Not -> "!"

(*helper function to translte char into string*)
let string_of_char =  
  String.make 1

(*translate all Literatls, 1. 3.4f "hello"*)
let string_of_literal = function 
  	IntL(i) -> string_of_int i
	|	FloatL(i) -> string_of_float i^"f"
	|	CharL(i) -> "\'" ^ string_of_char i ^ "\'"
	|	StringL(i) -> "\"" ^ i ^ "\""
	|	BoolL(i) ->string_of_bool i

(*translate return types of taml*)
let string_of_type_return = function 
    RInt -> "Integer"
  | RFloat ->"Float"
  | RChar  ->"Character"
  | RString ->"String"
  | RBoolean ->"Boolean"
   (*TaML exclusive*)
  | RTable ->"Table"
  | RLine ->"Line"
  | RCell ->"Cell"
  | RVoid ->"void"

(*translate tyeps*)
let string_of_type =function 
    Int ->"Integer"
  | Float ->"Float"
  | Char ->"Character"
  | String ->"String"
  | Boolean ->"Boolean"
   (*TaML exclusive*)
  | Table ->"Table"
  | Line ->"Line"
  | Cell ->"Cell"

(* translate expr into java code, this function is a helper function *)
(* of string_of_exprs which take Sast.ExprType as the input*)	
let rec string_of_expr = function 
  | Literal(l) -> string_of_literal l
  | TableAsCellValue(s,e1,e2) -> s^".getCellValue("^string_of_expr e1^","^string_of_expr e2^")"
  | LineAsCellValue(s,e) -> s^".getCellValue("^string_of_expr e^")"
  | CellAsCell(s) -> s^".getVal()"
  | TableAsCell(s,e1,e2) -> s^".getCell("^string_of_expr e1^","^string_of_expr e2^")"
  | LineAsCell(s,e) -> s^".getCell("^string_of_expr e^")"
  | Id(s) -> s
  | Uniop(op, e) -> string_of_operator1(op) ^ "(" ^ string_of_expr e ^ ")"
  | Binop(e1, o, e2) ->string_of_expr e1 ^ " " ^ string_of_operator(o) ^ " " ^ string_of_expr e2  
  |Assign(name,TableAsTable(s,e1,e2,e3,e4)) (*all table related assignment needs to be translated independently*)
		->	string_of_expr name^"="^s^".createTableCopy("^ string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e4^")"
  |Assign(name,TableAsLine_R(s,e1,e2,e3))
		->	string_of_expr name^"= new Line();\n "^string_of_expr name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^")"
  |Assign(name,TableAsLine_L(s,e1,e2,e3))
		->	string_of_expr name^"= new Line();\n "^string_of_expr name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e3^")"
  |Assign(name,TableAsCell(s,e1,e2))
		->	string_of_expr name^"="^s^".getCell("^string_of_expr e1^","^string_of_expr e2^")"
  |Assign(name,LineAsCell(s,e1))
		->	string_of_expr name^"="^s^".getCell("^string_of_expr e1^")"
  |Assign(TableAsCellValue(s,e1,e2),name)
		->	s^".setVal("^string_of_expr e1^","^string_of_expr e2^","^string_of_expr name^")"
  |Assign(TableAsTable(s,e1,e2,e3,e4),name)
		->	s^".setVal("^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e4^","^string_of_expr name^")"
  |Assign(TableAsLine_R(s,e1,e2,e3),name)
		->	s^".setVal("^string_of_expr e1^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr name^")"
  |Assign(TableAsLine_L(s,e1,e2,e3),name)
		->	s^".setVal("^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e3^","^string_of_expr name^")"
  |Assign(LineAsCellValue(s,e),name)
		 -> s^".setVal("^string_of_expr e^","^string_of_expr name^")"
  |Assign(LineAsLine(s,e1,e2),name)
 		-> s^".setVal("^string_of_expr e1^","^string_of_expr e2^","^string_of_expr name^")"
  |Assign(NewTotalCell(s),name)
		->	s^".setVal("^string_of_expr name^")"
  | Assign(v, e) ->	string_of_expr v ^ " = " ^ string_of_expr e
  | Call(f_Name,f_List) -> f_Name ^ "(" ^ String.concat "," (List.map string_of_expr f_List) ^ ")"
  | UnaryMinus (l)	->	"-"^string_of_expr l
  | All		->	"\"" ^ "ALL" ^ "\""
  | Null	->	"null"
  | _			->	"never get here!"

(*helper function that do explict casting on table related binary operations*)
let string_of_ascell expr tp = 
  if (tp = RInt) || (tp = RFloat) then 
  	match expr with
  		CellAsCell(i) -> "("^string_of_type_return tp^")"^ i^".getVal()"
  	| TableAsCellValue(s,e1,e2) -> "(" ^ (string_of_type_return tp) ^ ") " ^s^".getCellValue("^string_of_expr e1^","^string_of_expr e2^")"
  	| LineAsCellValue(s,e1) ->  "(" ^ (string_of_type_return tp) ^ ") " ^s^".getCellValue("^string_of_expr e1^")"
  	| _ as e1 -> string_of_expr e1
  else 
    string_of_expr expr
  
(*translate the expression from Sast, table related assignment needs to be translated independently*)
let rec string_of_exprs =	function  
  	ExprType(  Binop(e1, o, e2),t,tp) -> (string_of_ascell e1 tp) ^ string_of_operator(o) ^ (string_of_ascell e2 tp)
  | ExprType(Assign(TableAsCellValue(s,e1,e2),expr),tp,tp2) -> s^".setVal("^string_of_expr e1^","^string_of_expr e2^","^"(" ^ (string_of_type_return tp2) ^ ") " ^string_of_expr expr^")"
  | ExprType(Assign(LineAsCellValue(s,e1),expr2),tp,tp2) -> s^".setVal("^string_of_expr e1^","^"(" ^ (string_of_type_return tp2) ^ ")"^ string_of_expr expr2^")"
  | ExprType(Assign(CellAsCell(s),expr),tp,tp2) -> s^".setVal("^"(" ^ (string_of_type_return tp2) ^ ")"^string_of_expr expr^")"
  | ExprType(Assign(name,TableAsCellValue(s,e1,e2)),tp,tp2) ->
  		string_of_expr name^"= (" ^ (string_of_type_return tp) ^ ") " ^s^".getCellValue("^string_of_expr e1^","^string_of_expr e2^")"
  | ExprType(Assign(name,LineAsCellValue(s,e1)),tp,tp2) ->
  		string_of_expr name^"= (" ^ (string_of_type_return tp) ^ ") " ^s^".getCellValue("^string_of_expr e1^")"
  | ExprType(Assign(name,CellAsCell(s)),tp,tp2) ->
  		string_of_expr name^"= (" ^ (string_of_type_return tp) ^ ") " ^s^".getVal()"
  |ExprType(Assign(name,TableAsLine_R(s,e1,e2,e3)),t,tp2) ->
			string_of_expr name^"= new Line<"^string_of_type_return tp2^">();\n "^string_of_expr name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^")"
  |ExprType(Assign(name,TableAsLine_L(s,e1,e2,e3)),t,tp2) ->
			string_of_expr name^"= new Line<"^string_of_type_return tp2^">();\n "^string_of_expr name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e3^")"
  |ExprType(Assign(name,TableAsCell(s,e1,e2)),t,tp2) ->
			string_of_expr name^"= new Cell<"^string_of_type_return tp2^">();\n "^string_of_expr name^"="^s^".getCell("^string_of_expr e1^","^string_of_expr e2^")"
  |ExprType(Assign(name,LineAsCell(s,e1)),t,tp2) ->
			string_of_expr name^"= new Cell<"^string_of_type_return tp2^">();\n "^string_of_expr name^"="^s^".getCell("^string_of_expr e1^")" 
  |ExprType(esp,_,_)	-> string_of_expr esp

(*translate initialization from Sast, table related init  needs to be translated independently*)
let string_of_init =function  
  	(NullInit(Char,s),_)->"Character " ^" "^ s
  | (NullInit(t,s),_)-> string_of_type t^" "^ s ^"=null"
 	|	(VarInit(Char,s,Null),_)->"Character " ^" "^ s
  |	(VarInit(Table,name,TableAsTable(s,e1,e2,e3,e4)),_) ->
		"Table "^name^"="^s^".createTableCopy("^ string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e4^")"
  |	(VarInit(Line,name,TableAsLine_R(s,e1,e2,e3)),RInt) ->
		"Line<Integer> "^name^"= new Line<Integer>();\n"^name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^")" 
  |	(VarInit(Line,name,TableAsLine_R(s,e1,e2,e3)),RFloat) ->
		"Line<Float> "^name^"= new Line<Float>();\n"^name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^")"
  |	(VarInit(Line,name,TableAsLine_L(s,e1,e2,e3)),RInt) ->
		"Line<Integer> "^name^"= new Line<Integer>();\n"^name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e3^")"
  |	(VarInit(Line,name,TableAsLine_L(s,e1,e2,e3)),RFloat) ->
		"Line<Float> "^name^"= new Line<Float>();\n"^name^".assignLine("^s^","^string_of_expr e1^","^string_of_expr e2^","^string_of_expr e3^","^string_of_expr e3^")"
  |	(VarInit(Line,name,LineAsLine(s,e1,e2)),RInt) ->
		"Line<Integer> "^name^"="^s^".createLineCopy("^string_of_expr e1^","^string_of_expr e2^")"
  |	(VarInit(Line,name,LineAsLine(s,e1,e2)),RFloat) ->
		"Line<Float> "^name^"="^s^".createLineCopy("^string_of_expr e1^","^string_of_expr e2^")"
  |	(VarInit(Cell,name,TableAsCell(s,e1,e2)),RInt) ->
		"Cell<Integer> "^name^"="^s^".getCell("^string_of_expr e1^","^string_of_expr e2^")"
  |	(VarInit(Cell,name,TableAsCell(s,e1,e2)),RFloat) ->
		"Cell<Float> "^name^"="^s^".getCell("^string_of_expr e1^","^string_of_expr e2^")"
  |	(VarInit(Cell,name,LineAsCell(s,e1)),RInt) ->
		"Cell<Integer> "^name^"="^s^".getCell("^string_of_expr e1^")"
  |	(VarInit(Cell,name,LineAsCell(s,e1)),RFloat) ->
		"Cell<Float> "^name^"="^s^".getCell("^string_of_expr e1^")"
  |	(VarInit(Cell,name,Literal(i)),t) ->
		"Cell<"^string_of_type_return t^"> "^name^"=new Cell<"^string_of_type_return t^">();\n"^name^".setVal("^string_of_literal i^")"
  |	(VarInit(Char,s,e),_) ->
  		if(string_of_expr e="\'"^"\'")then  "char"^" "^s
			else "char"^" "^s^"="^string_of_expr e
  |	(VarInit(t,s,e),_) -> 
		string_of_type t^" "^s^"="^string_of_expr e

  |	(TableInit(t,s,e1,e2,taty),_)->"Table<"^string_of_type taty^"> "^ s^" = new Table<"^string_of_type taty^">("^string_of_expr e1^","^string_of_expr e2^")"

(*wrapper function that takes Sast.init as input and pass the parameters to string_of_init*)
let rec string_of_inits =function  
	SInit(i,t)-> string_of_init (i,t)

(*translation of statement*)
let rec string_of_stmt = function 
    SBlock(stmts) ->"{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | SExpr(expr) -> string_of_exprs expr ^ ";\n"
  | SReturn(e) -> "return " ^ string_of_exprs e ^ ";\n"
  | SPrint (e) -> "Printers.print("^string_of_exprs e^");\n"
  | SWhile(e, s) -> "while (" ^ string_of_exprs e ^ ")\n" ^ string_of_stmt s
  | SNoElseMark -> "{;}\n"
  | SBreak -> "break;\n"
  | SIf(e, s1, SNoElseMark) ->  "if (" ^ string_of_exprs e ^ ")\n" ^
      string_of_stmt s1
  | SIf(e, s1, s2) ->  "if (" ^ string_of_exprs e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | SFor(e1, e2, e3, stmt) ->"for (" ^string_of_exprs e1  ^ ";" ^string_of_exprs e2 ^ ";" ^ string_of_exprs e3 ^ ")\n" ^ string_of_stmt stmt ^ "\n"
  | SVarDecl(l)-> string_of_inits l^ ";\n"

(*translation of variable signature*)
let string_of_varsig =function 
	VarSig (t, id) ->string_of_type t ^ " " ^ id

(*translation of function signature*)
let string_of_funcsig=function 
	FuncSig(t, id) ->string_of_type_return t ^ " " ^ id

(*translation of initialzation*)
let string_of_global vdecl = string_of_init vdecl^ ";\n" 
 
(*function translation, main function translated independently*)
let string_of_func_decl func_decl = 
 	if(string_of_funcsig func_decl.sfname="void main") then
		default_keyword ^ " " ^string_of_funcsig func_decl.sfname ^ "(" ^ 
		"String[] args"	^ ")\n" ^ 
		string_of_stmt func_decl.sbody
 	else
		default_keyword ^ " " ^string_of_funcsig func_decl.sfname ^ "(" ^ 
		(String.concat "," (List.map string_of_varsig func_decl.sformals))
		^ ")\n" ^ 
	 	string_of_stmt func_decl.sbody

(*import list in java*)
let string_of_imports import_list =   
  String.concat ";\n" import_list ^ ";\n"

(*translation of constructs*)
let string_of_construct =function 
   	SGlobalVar(l)->"public static "^string_of_inits l ^ ";\n"
  |	SFuncDef (f)->string_of_func_decl f


(*translation of construct list and the generate whole java code*) 
let string_of_class x import_list programs =
	"\n" ^ string_of_imports import_list ^ "\n" ^
	"\npublic class " ^ x^ "\n{\n" ^String.concat "\n" (List.map string_of_construct programs) ^ "}\n"
