(* scanner_utils.ml - This utility is used for testing purposes.     *)
(* It outputs the results of the scanner (tokens) into string form   *)


open Scanner
open Parser

(* generate a string for each token. *)
let token_to_string token = 
   (match token with    
      | SEMI ->													(* Punctuation *)
           "SEMICOLON\n"                
      | LPAREN ->
           "LEFT_PAREN\n"
      | RPAREN ->
           "RIGHT_PAREN\n"
      | LBRACE ->
           "LEFT_BRACE\n"
      | RBRACE ->
           "RIGHT_BRACE\n"
      | COMMA ->
           "COMMA\n"
      | PLUS ->													(* Arithmetic *)
           "PLUS\n"											
      | MINUS ->
           "MINUS\n"
      | TIMES ->
           "TIMES\n"
      | DIVIDE  ->
           "DIVIDE\n"
      | MOD  ->
           "MOD\n"					
      | ASSIGN ->												(* Assignment *)
           "ASSIGN\n"
      | EQ ->														(* Comparison Logic *)
           "EQUAL\n"									
      | NEQ ->
           "NOT_EQUAL\n"	
      | REF ->"REFERENCE\n"
      | LT ->
           "LESS_THAN\n"
      | LEQ ->
           "LESS_THAN_OR_EQUAL\n"
      | GT  ->
           "GREATER_THAN\n"
      | GEQ ->
           "GREATER_THAN_OR_EQUAL\n"
      | AND ->													(* Boolean Logic *)
           "AND\n"
      | NOT ->
           "NOT\n"
      | OR ->
           "OR\n"
      | RETURN ->												(* Control Flow*)
           "RETURN\n"
      | IF ->
           "IF\n"
      | ELSE ->
           "ELSE\n"
      | FOR ->
           "FOR\n"
      | WHILE ->
           "WHILE\n"
	    | BREAK ->
           "BREAK\n"				
      | INT ->													(* Data Types *)
           "INT\n"
      | FLOAT ->
           "FLOAT\n"
      | BOOL ->
           "BOOL\n"
      | CHAR ->
           "CHAR\n"
      | STRING ->
           "STRING\n"
      | NULL ->													(* Null *)
           "NULL\n"
			| VOID ->                         (* Void *)
					 "VOID\n"
			| PRINT ->												(* TaML Exclusive *)
					 "PRINT\n"
      | LSQPAREN  ->										
           "LEFT_SQUARE_PAREN\n"
      | RSQPAREN  ->
           "RIGHT_SQUARE_PAREN\n"
      | WILDCARD ->
           "WILDCARD\n"
      | RANGE ->
           "RANGE\n"
      | CELL ->
           "CELL\n"
      | LINE ->
           "LINE\n"
      | TABLE ->
           "TABLE\n"
      | FUNC ->
           "FUNC\n"
      | INTLITERAL i ->										(* Literals/Constants *)
           "INTLITERAL: " ^ string_of_int(i) ^ "\n"
      | FLOATLITERAL f ->
           "FLOATLITERAL: " ^ string_of_float(f) ^ "\n"
      | CHARLITERAL c ->
           "CHARLITERAL: " ^ (Char.escaped c) ^ "\n"
      | STRINGLITERAL s ->
           "STRINGLITERAL: " ^ s ^ "\n"
      | BOOLLITERAL b ->
           "BOOLLITERAL: " ^ string_of_bool(b) ^ "\n"
      | ID id ->													(* Variables *)
           "ID: " ^ id ^ "\n"	
      | EOF ->														(* End of File *)
           "END_OF_FILE\n"	)


(* Generate a string for a set of tokens *)
let string_of_tokens tokens =
   let token_list = List.map token_to_string tokens in
   String.concat "" token_list
